/*
	TaltOS
	on the Philips Chip LPC2106

	By Peter Eordegh
*/

#include "lpc2106_memorymap.h"
#include "lpc2106_functions.h"
#include "TaltOS_defs.h"
#include "TaltOS.h"


void TaltOS_clear_tasklist()
{
	TaltOS_task* temp;
	unsigned int i;
	for(i = 0; i < TALTOS_TASK_LIST_SIZE; i++)
	{
		temp = ((TaltOS_task*)(TALTOS_TASK_LIST_PTR)) + i;
		temp->address  = TALTOS_TASK_NOADDRESS;
		temp->priority = TALTOS_TASK_NULL_PRIORITY;
		temp->state    = TALTOS_TASK_OFF;
		temp->argument = TALTOS_TASK_NULL_ARGUMENT;
	}
}

//Clear the flash operation buffer
void TaltOS_clear_flash_op_buffer()
{
	unsigned int i;
	for(i = 0; i < (TALTOS_FLASH_OP_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR + i)) = TALTOS_EMPTY;
	}
}


//Clear the input buffer
void TaltOS_clear_input_buffer()
{
	unsigned int i;
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	for(i = 0; i < (TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_INPUT_BUFFER_PTR + i)) = TALTOS_EMPTY;
	}
	env->input_buf_start = 0;
	env->input_buf_end = 0;
	env->input_buf_act_size = 0;
}

//Clear the output buffer
void TaltOS_clear_output_buffer()
{
	unsigned int i;
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	for(i = 0; i < (TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_INPUT_BUFFER_PTR + i)) = TALTOS_EMPTY;
	}
	env->output_buf_start = 0;
	env->output_buf_end = 0;
	env->output_buf_act_size = 0;
}

//Clear the environement variables of the TalOS
void TaltOS_clear_environement()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	env->input_buf_start	= 0;
	env->input_buf_end		= 0;
	env->input_buf_act_size = 0;
	env->output_buf_start	= 0;
	env->output_buf_end		= 0;
	env->output_buf_act_size= 0;

	env->actual_task = TALTOS_TASK_NULL_PTR;
	
	env->TaltOS_starting_address = TaltOS_get_component_address(TALTOS_ID1, TALTOS_ID2);
}

//The TaltOS initialization routine
void TaltOS_init()
{
	DEBUG('T')
	TaltOS_clear_flash_op_buffer();
	TaltOS_clear_input_buffer();
	TaltOS_clear_output_buffer();
	TaltOS_clear_environement();
	TaltOS_clear_tasklist();
	//LPC2106 specific code
	REG8(TIMER0_TCR) = 0x1;
	lpc2106_vectored_interrupts_setup();
	//End of LPC2106 specific code
	TaltOS_run_and_list_components(TALTOS_RUN_AND_LIST_SHOULDLIST | TALTOS_RUN_AND_LIST_SHOULDRUN);
	TaltOS_loop();
}

//The TaltOS main loop
void TaltOS_loop()
{
	while(1)
	{
		TaltOS_set_run_end_next_task();
	}
}

//Searches, runs and ends the next task in the task list (concerning prioroty)
void TaltOS_set_run_end_next_task()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	TaltOS_task* temp_task;
	unsigned int i;
	for(i = 0; i < TALTOS_TASK_LIST_SIZE; i++)
	{
		temp_task = (TaltOS_task*)TALTOS_TASK_LIST_PTR + i;
		if (temp_task->state == TALTOS_TASK_READYTORUN)
		{
			if (env->actual_task == TALTOS_TASK_NULL_PTR)
			{
				env->actual_task = temp_task;
			}
			else
			{
				if (env->actual_task->priority > temp_task->priority)
				{
					env->actual_task = temp_task;
				}
			}
		}
	}
	
	if (!env->actual_task == TALTOS_TASK_NULL_PTR)
	{
		void (*task)(unsigned long) = (void(*)(unsigned long))env->actual_task->address;
		// Set the task's state to running
		env->actual_task->state = TALTOS_TASK_RUNNING;
		// Run the actual task
		task(env->actual_task->argument);
		// Ending of the actual task
		env->actual_task->state    = TALTOS_TASK_OFF;
		env->actual_task->address  = TALTOS_TASK_NOADDRESS;
		env->actual_task->priority = TALTOS_TASK_NULL_PRIORITY;
		env->actual_task->argument = TALTOS_TASK_NULL_ARGUMENT;
		env->actual_task = TALTOS_TASK_NULL_PTR;
	}
}

//The Task adding function
unsigned int TaltOS_add_task(unsigned long address, unsigned short priority, unsigned long argument)
{
	TaltOS_task* temp_task;
	unsigned int i;
	for(i = 0; i < TALTOS_TASK_LIST_SIZE; i++)
	{
		temp_task = (TaltOS_task*)TALTOS_TASK_LIST_PTR + i;
		if (temp_task->state == TALTOS_TASK_OFF)
		{
			temp_task->address  = address;
			temp_task->priority = priority;
			temp_task->argument = argument;
			temp_task->state    = TALTOS_TASK_READYTORUN;
			return TALTOS_TRUE;
		}
	}
	return TALTOS_FALSE;
}

//Run and list all the components' init function (needed after a reset)
void TaltOS_run_and_list_components(unsigned long should_do)
{
	unsigned int i;
	unsigned int should_run = 0;
	unsigned int should_list = 0;
	TaltOS_reg_entity *comp_reg;
	void (*comp_init)();
	
	if (should_do & TALTOS_RUN_AND_LIST_SHOULDRUN)
	{
		should_run = 1;
	}
	if (should_do & TALTOS_RUN_AND_LIST_SHOULDLIST)
	{
		should_list = 1;
	}
	
	for(i = 0; i < TALTOS_COMPONENT_TABLE_SIZE / sizeof(TaltOS_reg_entity); i++)
	{
		comp_reg = (TaltOS_reg_entity*)TALTOS_COMPONENT_TABLE_PTR + i;
		if((comp_reg->ID1 != TALTOS_EMPTY) && (comp_reg->ID2 != TALTOS_EMPTY))
		{
			if (should_list)
			{
				TaltOS_long_output_write(TALTOS_CRLFCRLF);
				TaltOS_long_output_write(comp_reg->ID1 + TALTOS_00000000);
				TaltOS_long_output_write(comp_reg->ID1 + TALTOS_00000000);
			}
			if (should_run)
			{
				if(	!((comp_reg->ID1 == TALTOS_ERASED) && (comp_reg->ID2 == TALTOS_ERASED)) &&
					!((comp_reg->ID1 == TALTOS_ID1) && (comp_reg->ID2 == TALTOS_ID2)) &&
					!(TaltOS_is_garbage_reg(comp_reg)))
				{
					comp_init = (void(*)())TaltOS_get_component_function(comp_reg->address, COMP_INIT);
					comp_init();
				}
			}
		}
	}
}


//The TaltOS tic routine for the preemtive task change
void TaltOS_tic()
{
	//REG8(UART0_THR) = 'p';
}

//The input buffer's input function, can be called by the microcontroller
void TaltOS_long_input_write(unsigned long inlong)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	unsigned long* address = (unsigned long*)(TALTOS_INPUT_BUFFER_PTR) + env->input_buf_end;
	unsigned long  comp_address;
	(*address) = inlong;

	if (env->input_buf_end < (TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)))
	{
		if (env->input_buf_start != env->input_buf_end + 1){env->input_buf_end++;}
	}
	else
	{
		if (env->input_buf_start != 0){env->input_buf_end = 0;}
	}

	(env->input_buf_act_size)++;

	comp_address = TaltOS_get_component_function(env->TaltOS_starting_address, TALTOS_CHECK_ORDER);
	
	if (!TaltOS_add_task(comp_address, TALTOS_OWN_TASK_PRIORITY, TALTOS_TASK_NULL_ARGUMENT))
	{
		TaltOS_long_output_write(TALTOS_TF_CRLF);
	}

}


//The input buffer's output function, can be called by the tasks to recive data from out of the system
unsigned int TaltOS_long_input_read(unsigned long* ret)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	if (env->input_buf_start == env->input_buf_end){return TALTOS_FALSE;}
	(*ret) = *(((unsigned long*)TALTOS_INPUT_BUFFER_PTR) + env->input_buf_start);

	if (env->input_buf_start < TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)) {env->input_buf_start++;}
	else {env->input_buf_start = 0;}
	(env->input_buf_act_size)--;
	return TALTOS_TRUE;
}

//The output buffer's input function, can be called by the tasks to send data out of system
void TaltOS_long_output_write(unsigned long inlong)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	unsigned long* address = (unsigned long*)(TALTOS_OUTPUT_BUFFER_PTR) + env->output_buf_end;
	(*address) = inlong;

	if (env->output_buf_end < TALTOS_OUTPUT_BUFFER_SIZE / sizeof(unsigned long))
	{
		if (env->output_buf_start != env->output_buf_end + 1){env->output_buf_end++;}
	}
	else
	{
		if (env->output_buf_start != 0){env->output_buf_end = 0;}
	}

	(env->output_buf_act_size)++;

	//LPC2106 specific code
	lpc2106_force_uart_irq();
	//LPC2106 specific code end
}

//The output buffer's output function, can be called by the microcontroller
unsigned int TaltOS_long_output_read(unsigned long* ret)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	if (env->output_buf_start == env->output_buf_end){return TALTOS_FALSE;}
	(*ret) = *((unsigned long*)TALTOS_OUTPUT_BUFFER_PTR + env->output_buf_start);

	if (env->output_buf_start < TALTOS_OUTPUT_BUFFER_SIZE / sizeof(unsigned long)) {env->output_buf_start++;}
	else {env->output_buf_start = 0;}
	(env->output_buf_act_size)--;
	return TALTOS_TRUE;
}


//Get the component's address by it's ID (Absolute address)
unsigned long* TaltOS_get_component_address(unsigned long comp_id1, unsigned long comp_id2)
{
	unsigned int isfound = 0;
	TaltOS_reg_entity *comp_reg;
	unsigned int i;
	for(i = 0; (i < TALTOS_COMPONENT_TABLE_SIZE/sizeof(TaltOS_reg_entity)) && (!isfound); i++)
	{
		comp_reg = (TaltOS_reg_entity*)TALTOS_COMPONENT_TABLE_PTR + i;
		if ((comp_reg->ID1 == comp_id1) && (comp_reg->ID2 == comp_id2))
		{
			isfound = 1;
		}
	}
	if (isfound) {return comp_reg->address;}
	else {return 0;}
}

//Get the component's n. function's pointer (Absolute address)
unsigned long TaltOS_get_component_function(unsigned long* address, unsigned long index)
{
	return (unsigned long)(address + (*(address + index)) / sizeof(unsigned long));
}

//Check if a component has arrived into the input buffer
void TaltOS_check_order()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	unsigned long comp_address;
	if(env->input_buf_act_size >= 1)
	{
		switch (*((unsigned long*)(TALTOS_INPUT_BUFFER_PTR) + env->input_buf_end - 1))
		{
			case (TALTOS_RUN_INIT)		:
			{
				TaltOS_end_order();
				comp_address = TaltOS_get_component_function(env->TaltOS_starting_address, TALTOS_RUN_AND_LIST_COMPONENTS);
				if (!TaltOS_add_task(comp_address, TALTOS_OWN_TASK_PRIORITY, TALTOS_RUN_AND_LIST_SHOULDRUN))
				{
					TaltOS_long_output_write(TALTOS_TF_CRLF);
				}
				break;
			}
			case (TALTOS_COMP_BEGIN)	: 
			{
				TaltOS_end_order();
				break;
			}
			case (TALTOS_COMP_END)	: 
			{
				TaltOS_end_order();
				break;
			}
			case (TALTOS_REM_COMP)		: 
			{
				TaltOS_end_order();
				break;
			}
			case (TALTOS_COMP_LIST)		: 
			{
				TaltOS_end_order();
				comp_address = TaltOS_get_component_function(env->TaltOS_starting_address, TALTOS_RUN_AND_LIST_COMPONENTS);
				if (!TaltOS_add_task(comp_address, TALTOS_OWN_TASK_PRIORITY, TALTOS_RUN_AND_LIST_SHOULDLIST))
				{
					TaltOS_long_output_write(TALTOS_TF_CRLF);
				}
				break;
			}
		}
	}
}

void TaltOS_end_order()
{
	TaltOS_long_output_write(TALTOS_OK_CRLF);
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	if (env->input_buf_end == 0) {env->input_buf_end = TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long);}
	else {env->input_buf_end--;}
	env->input_buf_act_size--;
}


//Checks if the component's ids are garbage (wrong value => error in erase operation)
unsigned int TaltOS_is_garbage_reg(TaltOS_reg_entity *test_reg)
{
	unsigned char* test_array = (unsigned char*)test_reg;
	if (test_array[0] == TALTOS_COMPID_END && test_array[4] == TALTOS_COMPID_END) {return 0;}
	else return 1;
}

void TaltOS_copy_component()
{
	/*unsigned long* empty_tester;
	int address_found = 0;
	unsigned long i = 0;
	unsigned char c;
	TaltOS_reg_entity component;
	TaltOS_reg_entity *TaltOS_reg_flash;
	TaltOS_reg_entity *TaltOS_reg_buffer;
	
	
	//Write the component to the flash operand buffer
	while(TaltOS_char_input_read(&c))
	{
		(*((unsigned char*)TALTOS_FLASH_OP_BUFFER_PTR + i)) = c;
		i++;
	}
	i--;
	//Get the ID's of the component
	component.ID1 = *((unsigned long*)(TALTOS_FLASH_OP_BUFFER_PTR + i-13))
	*((unsigned long*)(TALTOS_FLASH_OP_BUFFER_PTR + i-13)) = TALTOS_EMPTY;
	component.ID2 = *((unsigned long*)(TALTOS_FLASH_OP_BUFFER_PTR + i-9))
	*((unsigned long*)(TALTOS_FLASH_OP_BUFFER_PTR + i-13)) = TALTOS_EMPTY;
	//Get the size of the component
	component.size = (i - 13 / LPC2106_MINDATA) + 1;

	//Search for the first empty place of the TaltOS
	for(i=0; (i<LPC2106_FLASH_MEMORY_SIZE/LPC2106_MINDATA) && (!address_found); i++)
	{
		address_found=true;
		// Check if the full size is fit to this place
		for(j=0; (j<TaltOS.size)&&(address_found); j++)
		{
			empty_tester=(unsigned long*)((i+j)*LPC2106_MINDATA);
			if ((*empty_tester)!=TALTOS_EMPTY)
			{
				address_found=false;
			}
		}
	}

	if (address_found)
	{
		TaltOS.address = (unsigned long*)((i-1)*LPC2106_MINDATA);

		//Copy the buffer by LPC2106_MINDATA size blocks
		for(j=0; j<TaltOS.size; j++)
		{
			copy_to_flash_512(	(unsigned long)TaltOS.address+j*LPC2106_MINDATA,
								(unsigned long*)(TALTOS_INPUT_BUFFER_PTR+j*LPC2106_MINDATA),
								1);
		}

		clear_input_buffer();

		//Save the full component table, and register the TaltOS if possible
		address_found=false;
		for(i = 0; i < (TALTOS_FLASH_OP_BUFFER_SIZE/sizeof(TaltOS_reg_entity)); i++)
		{
			(*((TaltOS_reg_entity*)(TALTOS_FLASH_OP_BUFFER_PTR)+i)) = (*((TaltOS_reg_entity*)(TALTOS_COMPONENT_TABLE_PTR)+i));
			TaltOS_reg = (TaltOS_reg_entity*)(TALTOS_FLASH_OP_BUFFER_PTR)+i;
			if ((!address_found)&&(TaltOS_reg->ID1 == TALTOS_EMPTY)&&(TaltOS_reg->ID2 == TALTOS_EMPTY))
			{
				TaltOS_reg->ID1=TaltOS.ID1;
				TaltOS_reg->ID2=TaltOS.ID2;
				TaltOS_reg->address=TaltOS.address;
				TaltOS_reg->size=TaltOS.size;
				address_found=true;
			}
		}

		//Copy the full component table back to the flash
		copy_to_flash_512(TALTOS_COMPONENT_TABLE_PTR,(unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR,16);

		clear_flash_op_buffer();

		//Run the TaltOS
		what = 1;
	}
	else
	{
		REG8(UART0_THR) = 'F';		//F as Fail
		what = 0;
	}*/
}
