/************************************************************************
* This functions are used for initializing LPC2106 chip on an eval board*
************************************************************************/

#include "lpc2106_memorymap.h"
#include "lpc2106_functions.h"
#include "TaltOS_defs.h"
#include "TaltOS.h"

typedef void (*IAP) (unsigned long[], unsigned long[]);

void timer0_irq_handler();
void default_irq_handler();
void uart0_irq_handler();

// Set the IT handlers on the LPC2106
void lpc2106_vectored_interrupts_setup()
{
	unsigned long* TaltOS_address = TaltOS_get_component_address(TALTOS_ID1, TALTOS_ID2);
	REG32(VICIntSelect) = 0x0;	// All interrupts are assigned to IRQ category
	REG32(VICIntEnable) = 0x50;	// Timer0 (bit4) uart0 (bit6) interrupt is enabled  /01010000/
	REG32(VICVectAddr0) = TaltOS_get_component_function(TaltOS_address, LPC2106_TIMER_HANDLER);	// Timer0 IRQ handler
	REG32(VICVectCntl0) = 0x24; // interrupt source is timer0 (source 4) and enable (bit 5) /00100100/
	REG32(VICVectAddr1) = TaltOS_get_component_function(TaltOS_address, LPC2106_UART_HANDLER);     // UART0 IRQ handler
	REG32(VICVectCntl1) = 0x26;  // interrupt source is uart0 (source 6) and enable (bit 5) /00100110/
	REG32(VICDefVectAddr) = TaltOS_get_component_function(TaltOS_address, LPC2106_DEFAULT_HANDLER);

	//Disable UART_THRE interrupt, it will be turned on if data avaible
	REG8(UART0_IER) = 0x01;

}

// Forces an uart THRE interrupt if it was disabled
void lpc2106_force_uart_irq()
{
	if (!(REG8(UART0_IER) & 0x02))
	{
		unsigned long temp;
		int j;
		TaltOS_long_output_read(&temp);
		REG8(UART0_IER) = 0x03;
		for(j = 0; j < sizeof(unsigned long); ++j){REG8(UART0_THR) = ((unsigned char*)(&temp))[j];}
	}
}

// The flash writing routine
long lpc2106_copy_to_flash_512(unsigned long to_address, unsigned long buffer[], long data_size)
{
	unsigned long command[5];								//Command pointer for IAP
	unsigned long result[2];								//Result pointer for IAP
	unsigned long vicmask = REG32(VICIntEnable);			//To save IT mas
	IAP iap_entry;											//The function pointer
	unsigned long sector = to_address / LPC2106_SECTOR_SIZE;//Calculate which sector we want to write
	REG32(VICIntEnClr) = 0xFFFFFFFF;						//Disable IT
	iap_entry=(IAP)IAP_LOCATION;							//Get IAP routine pointer

	command[0] = 50;										//Prepare sector to write/erase
	command[1] = sector;									// from sector
	command[2] = sector;									// till sector
	iap_entry(command, result);								//launch IAP
	//	REG8(UART0_THR) = result[0]+'0';

	command[0] = 51;										//Write sector
	command[1] = to_address - (to_address % 512);			//Address on correct boundary (512);
	command[2] = (unsigned long)buffer;						//Buffer start
	command[3] = LPC2106_MINDATA * data_size;					//Buffer size
	//Processor clock frequenci in kHz (External cristal:10 MHz M=6)
	command[4] = 60000;
	iap_entry(command, result);								//launch IAP

	//	REG8(UART0_THR) = result[0]+'0';
	REG32(VICIntEnable) = vicmask;							//Enable IT
	return result[0];
}

// Timer interrupt
void lpc2106_timer0_irq_handler()
{
	TaltOS_tic();
	REG8(TIMER0_IR) = 0xFF;		// Clear the interrupt flag
}

//Default interrupt handler
void lpc2106_default_irq_handler()
{
	REG8(UART0_THR) = '@';
}

//UART interrupt handler
void lpc2106_uart0_irq_handler()
{

	switch (REG8(UART0_IIR) & 0x0E) //Mask with 1110
	{
		case (0x02) :				//In case of THRE interrupt (more often interrupt)
		{
			unsigned long temp;
			int i;
			int j;
			
			for(i = 0; (i < 4) && (TaltOS_long_output_read(&temp)); ++i)
			{
				for(j = 0; j < sizeof(unsigned long); ++j){REG8(UART0_THR) = ((unsigned char*)(&temp))[j];}
			} //Write out max 16 bytes (this is the size of the THR which is empty)
			
			if (i == 0) {REG8(UART0_IER) = 0x01;} //If we had nothing to write out

			break;
		}
		case (0x04) :				//In case of RDA interrupt (less often interrupt)
		{
			unsigned char temp[4];
			int i;
			for(i = 0; i < 4; i++) {temp[i] = REG8(UART0_RBR);}
			TaltOS_long_input_write(*((unsigned long*)(temp)));
			break;
		}
	}
}


