/*
	Bootloader for the TaltOS
	on the Philips Chip LPC2106

	By Peter Eordegh
*/


#include "lpc2106_init.h"
#include "lpc2106_memorymap.h"
#include "TaltOS_bootloader_defs.h"

int buffer_pointer;
int what;
unsigned long reciving_size;

long copy_to_flash_512(unsigned long to_address, unsigned long buf[], unsigned long data_size);

//Clear the input buffer
void clear_input_buffer()
{
	unsigned int i;
	for(i = 0; i < (TALTOS_INPUT_BUFFER_SIZE/sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_INPUT_BUFFER_PTR + i)) = TALTOS_EMPTY;
	}
	buffer_pointer = 0;
}

//Clear the flash operation buffer
void clear_flash_op_buffer()
{
	unsigned int i;
	for(i = 0; i < (TALTOS_FLASH_OP_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR + i)) = TALTOS_EMPTY;
	}
}

unsigned int is_garbage_reg(unsigned char* test_array)
{
	if (test_array[0] == 0x4E && test_array[4] == TALTOS_COMPID_END) {return 0;}
	else return 1;
}


//Remove all TaltOS registry entities
void remove_TaltOS()
{

	TaltOS_reg_entity *TaltOS_reg_flash;
	TaltOS_reg_entity *TaltOS_reg_buffer;
	unsigned int i;

	//Check the component table, if it is a TaltOS, write the complement if it's ID to the buffer
	for(i = 0; i < (TALTOS_FLASH_OP_BUFFER_SIZE / sizeof(TaltOS_reg_entity)); i++)
	{
		TaltOS_reg_flash = (TaltOS_reg_entity*)(TALTOS_COMPONENT_TABLE_PTR) + i;
		if ((TaltOS_reg_flash->ID1 == TALTOS_ID1) && (TaltOS_reg_flash->ID2 == TALTOS_ID2))
		{
			TaltOS_reg_buffer = (TaltOS_reg_entity*)(TALTOS_FLASH_OP_BUFFER_PTR) + i;
			TaltOS_reg_buffer->ID1= ~TALTOS_ID1;
			TaltOS_reg_buffer->ID2= ~TALTOS_ID2;
			// We don't erase the address and size, for later garbage collection
		}
		else if ((TaltOS_reg_flash->ID1 != TALTOS_EMPTY)&&(TaltOS_reg_flash->ID2 != TALTOS_EMPTY))
		{
			if (is_garbage_reg((unsigned char*)TaltOS_reg_flash))
			{
				TaltOS_reg_buffer = (TaltOS_reg_entity*)(TALTOS_FLASH_OP_BUFFER_PTR) + i;
				TaltOS_reg_buffer->ID1= ~TaltOS_reg_flash->ID1;
				TaltOS_reg_buffer->ID2= ~TaltOS_reg_flash->ID1;
			}
		}
	}

	copy_to_flash_512(TALTOS_COMPONENT_TABLE_PTR,(unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR,16);

	clear_flash_op_buffer();

	DEBUG('R') // R as Removed
	
	what = 0;
}

//Run the TaltOS
void run_TaltOS()
{
	unsigned int i;
	int haverun = 0;
	TaltOS_reg_entity *TaltOS_reg;
	for(i = 0; (i < TALTOS_COMPONENT_TABLE_SIZE / sizeof(TaltOS_reg_entity)) && (!haverun); i++)
	{
		TaltOS_reg = (TaltOS_reg_entity*)TALTOS_COMPONENT_TABLE_PTR + i;
		if ((TaltOS_reg->ID1 == TALTOS_ID1) && (TaltOS_reg->ID2 == TALTOS_ID2))
		{
			unsigned long* rel_ptr = (unsigned long*)(TaltOS_reg->address);
			void (*TaltOS)() =(void(*)())(TaltOS_reg->address + (*rel_ptr) / 4);
			TaltOS();
			haverun = 1;
		}
	}
	if (!haverun){DEBUG('N');} // N as Not Found

	clear_input_buffer();

	what = 0;
}

void copy_and_run_TaltOS()
{
	unsigned long* empty_tester;
	int address_found = 0;
	unsigned long i,j;
	TaltOS_reg_entity TaltOS;
	TaltOS_reg_entity *TaltOS_reg_flash;
	TaltOS_reg_entity *TaltOS_reg_buffer;
	TaltOS.ID1 = TALTOS_ID1;
	TaltOS.ID2 = TALTOS_ID2;
	//Get the size of the buffer to be written
	TaltOS.size = (buffer_pointer * sizeof(unsigned long)/LPC2106_MINDATA) + 1;

	//Search for the first empty place of the TaltOS
	for(i = 0; (i < LPC2106_FLASH_MEMORY_SIZE / LPC2106_MINDATA) && (!address_found); i++)
	{
		address_found = 1;
		// Check if the full size is fit to this place
		for(j = 0; (j < TaltOS.size) && (address_found); j++)
		{
			empty_tester = (unsigned long*)((i + j)*LPC2106_MINDATA);
			if ((*empty_tester) != TALTOS_EMPTY)
			{
				address_found = 0;
			}
		}
	}

	if (address_found)
	{
		TaltOS.address = (unsigned long*)((i - 1) * LPC2106_MINDATA);
		
		//Copy the buffer by LPC2106_MINDATA size blocks
		for(j = 0; j < TaltOS.size; j++)
		{
			copy_to_flash_512(	(unsigned long)TaltOS.address + j * LPC2106_MINDATA,
								(unsigned long*)(TALTOS_INPUT_BUFFER_PTR + j * LPC2106_MINDATA),
								1);
		}


		clear_input_buffer();

		//Register the TaltOS if possible
		address_found = 0;
		for(i = 0; (i < (TALTOS_FLASH_OP_BUFFER_SIZE / sizeof(TaltOS_reg_entity))) && (!address_found); i++)
		{
			TaltOS_reg_flash = ((TaltOS_reg_entity*)TALTOS_COMPONENT_TABLE_PTR) + i;
			if ((TaltOS_reg_flash->ID1 == TALTOS_EMPTY) && (TaltOS_reg_flash->ID2 == TALTOS_EMPTY))
			{
				TaltOS_reg_buffer = ((TaltOS_reg_entity*)TALTOS_FLASH_OP_BUFFER_PTR) + i;
				TaltOS_reg_buffer->ID1 = TaltOS.ID1;
				TaltOS_reg_buffer->ID2 = TaltOS.ID2;
				TaltOS_reg_buffer->address = TaltOS.address;
				TaltOS_reg_buffer->size = TaltOS.size;
				address_found = 1;
			}
		}


		//Copy the full component table back to the flash
		copy_to_flash_512(	TALTOS_COMPONENT_TABLE_PTR,
							(unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR,
							LPC2106_SECTOR_SIZE / LPC2106_MINDATA);

		clear_flash_op_buffer();
		

		//Run the TaltOS
		what = 1;
	}
	else
	{
		DEBUG('F');		//F as Fail
		what = 0;
	}
}


//Copy 512 bytes to the Flash adress given
long copy_to_flash_512(unsigned long to_address, unsigned long buf[], unsigned long data_size)
{
	unsigned long command[5];								//Command pointer for IAP
	unsigned long result[2];								//Result pointer for IAP
	unsigned long vicmask = REG32(VICIntEnable);			//To save IT mas
	IAP iap_entry;											//The function pointer
	unsigned long sector = to_address / LPC2106_SECTOR_SIZE;//Calculate which sector we want to write
	REG32(VICIntEnClr) = 0xFFFFFFFF;						//Disable IT
	iap_entry=(IAP) IAP_LOCATION;							//Get IAP routine pointer

	command[0] = 50;										//Prepare sector to write/erase
	command[1] = sector;									// from sector
	command[2] = sector;									// till sector
	iap_entry(command,result);								//launch IAP
	//	DEBUG(result[0]+'0')

	command[0] = 51;										//Write sector
	command[1] = to_address - (to_address % 512);			//Address on correct boundary (512);
	command[2] = (unsigned long)buf;						//Buffer start
	command[3] = LPC2106_MINDATA * data_size;					//Buffer size
	//Processor clock frequenci in kHz (External cristal:10 MHz M=6)
	command[4] = 60000;
	iap_entry(command,result);								//launch IAP

	//	DEBUG(result[0]+'0')
	REG32(VICIntEnable) = vicmask;							//Enable IT
	return result[0];
}

void timer0_irq_handler()
{
	REG8(UART0_THR) = 't';
	REG8(TIMER0_IR) = 0xFF;		// Clear the interrupt flag
}

//Clear order from the end of the buffer
void end_order()
{
	buffer_pointer -= 1;
	*((unsigned long*)(TALTOS_INPUT_BUFFER_PTR) + buffer_pointer) = TALTOS_EMPTY;
}

// Chack if there is an order
unsigned int order(void)
{
	unsigned int ret = 0;
	unsigned long check;
	if (buffer_pointer >= 1)
	{
		check = *((unsigned long*)(TALTOS_INPUT_BUFFER_PTR) + buffer_pointer - 1);
		switch (what)
		{
			case (0) :
			{
				switch (check)
				{
					case (TALTOS_RUN_INIT)		: 
					{
						end_order();
						ret = 1;
						break;
					}
					case (TALTOS_REM_COMP)		: 
					{
						end_order();
						ret = 3;
						break;
					}
					case (TALTOS_COMP_BEGIN)	: 
					{
						end_order();
						ret = 4;
						break;
					}
				}
				break;
			}
			case (4) : 
			{
				reciving_size = check;
				end_order();
				ret = 5;
				break;
			}
			case (5) :
			{
				if ((check == TALTOS_COMP_END) && 
					((reciving_size / sizeof(unsigned long)) == buffer_pointer))
				{
					end_order();
					ret = 2;
				}
				else {ret = 5;}
				break;
			}
		}
	}
	return ret;
}

//Default interrupt handler
void default_irq_handler(void)
{
	REG8(UART0_THR) = '@';
}

//UART interrupt handler
void uart0_irq_handler(void)
{
	switch (REG8(UART0_IIR) & 0x0E) //Mask with 1110
	{
		case (0x02) :				//In case of THRE interrupt (more often interrupt)
		{
			REG8(UART0_IER) = 0x01; //If we have nothing to write out
			break;
		}
		case (0x04) :				//In case od RDA interrupt (less often interrupt)
		{
			unsigned char temp[4];
			int i;
			for(i = 0; i < 4; i++)
			{
				temp[i] = REG8(UART0_RBR);
			}
			*((unsigned long*)(TALTOS_INPUT_BUFFER_PTR) + buffer_pointer++) = *((unsigned long*)(temp));
			what = order();	//Check if there was an order or not
			break;
		}
	}
}

int main (void)
{
	what = 0;
	lpc2106_run();		//Init chip
	REG8(UART0_THR) = 'G';		//Write 'G' for Go
	REG8(TIMER0_TCR) = 0x0;		//Disable Timer interrupt
	clear_input_buffer();				//Clear the input buffer
	clear_flash_op_buffer();			//Clear the input buffer
	lpc2106_vectored_interrupts_init(	(unsigned long) timer0_irq_handler,
										(unsigned long) uart0_irq_handler,
										(unsigned long) default_irq_handler); //Set interrupt handlers
	if (!(REG8(UART0_IER) & 0x02)){REG8(UART0_IER) = 0x03;}
	while(1)	//Loop for order test
	{
		switch (what)
		{
			case (0) :  break;
			case (1) :  run_TaltOS();break;
			case (2) :  copy_and_run_TaltOS();break;
			case (3) :  remove_TaltOS();break;
		}
	}
}
